@@ -1,4 +1,4 @@
-/* $Id: gdict-defbox.c,v 1.35 2005/01/05 14:27:58 vnoel Exp $ */
+/* $Id: gdict-defbox.c,v 1.41 2005/01/24 21:17:27 vnoel Exp $ */
 
 /*
  *  Mike Hughes <mfh@psilord.com>
@@ -40,17 +40,227 @@ static gint gdict_defbox_signals[LAST_SIGNAL] = { 0, 0, 0, 0 };
 static void gdict_defbox_init (GDictDefbox *defbox);
 static void gdict_defbox_class_init (GDictDefboxClass *class);
 
-static void def_error_cb  (dict_command_t *command, DictStatusCode code, 
-                           gchar *message, gpointer data);
-static void def_status_cb (dict_command_t *command, DictStatusCode code, 
-                           int num_found, gpointer data);
-static void def_data_cb   (dict_command_t *command, dict_res_t *res,
-                           gpointer data);
-
+static void gdict_defbox_error_cb  (dict_command_t *command, DictStatusCode code, 
+                                    gchar *message, gpointer data);
+static void gdict_defbox_status_cb (dict_command_t *command, DictStatusCode code, 
+                                    int num_found, gpointer data);
+static void gdict_defbox_data_cb   (dict_command_t *command, dict_res_t *res,
+                                    gpointer data);
 static gboolean is_xref   (gchar *str, int len);
 static gboolean is_number (gchar *str, int len);
 static gboolean is_part   (gchar *str, int len);
-static void defbox_add    (GDictDefbox *defbox, gchar *def);
+
+/* Setup textview tags
+ */
+void
+gdict_defbox_setup_tags (GDictDefbox *defbox)
+{
+    GtkTextBuffer *buffer;
+	
+    buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (defbox));
+	
+    gtk_text_buffer_create_tag (buffer, "italic",
+			        "style", PANGO_STYLE_ITALIC, NULL);
+    gtk_text_buffer_create_tag (buffer, "bold",
+			        "weight", PANGO_WEIGHT_BOLD, NULL);
+    gtk_text_buffer_create_tag (buffer, "underline",
+			        "underline", PANGO_UNDERLINE_SINGLE, NULL);
+}
+
+static void
+gdict_defbox_lookup_start_cb (GtkWidget *widget, gpointer data) 
+{
+  GDictWindow *gdict = data;
+  static GdkCursor *busy_cursor = NULL;
+  GtkWidget *window_widget = gtk_widget_get_toplevel(GTK_WIDGET(gdict));
+
+  if (busy_cursor == NULL)
+    busy_cursor = gdk_cursor_new (GDK_WATCH);
+  gdk_window_set_cursor (GTK_WIDGET(gdict)->window, busy_cursor);
+  gdict_menus_set_sensitive (gdict, FALSE);
+
+  gtk_window_set_title (GTK_WINDOW (gdict), _("Dictionary"));
+  gtk_statusbar_push (GTK_STATUSBAR (gdict->statusbar), 0, _("Looking up word..."));
+}
+
+static void
+gdict_defbox_lookup_done_cb (GtkWidget *widget, gpointer data) 
+{
+  gchar *word, *title;
+  GDictWindow *gdict = data;
+  GtkAction *action;
+
+  word = gtk_editable_get_chars (GTK_EDITABLE (gdict->word_entry), 0, -1);
+  title = g_strconcat (word, " - ", _("Dictionary"), NULL);
+  gtk_window_set_title (GTK_WINDOW (gdict), title);
+  g_free (word);
+  g_free (title);
+
+  gtk_statusbar_pop (GTK_STATUSBAR (gdict->statusbar), 0);
+  gtk_statusbar_push (GTK_STATUSBAR (gdict->statusbar), 0, _("Lookup done"));
+  gdict_menus_set_sensitive (gdict, TRUE);
+
+  gdk_window_set_cursor (GTK_WIDGET(gdict)->window, NULL);
+}
+
+static void
+gdict_defbox_not_found_cb (GtkWidget *widget, gpointer data) 
+{
+  GDictWindow *gdict = data;
+  gchar *word;
+  gtk_statusbar_pop (GTK_STATUSBAR (gdict->statusbar), 0);
+  gtk_statusbar_push (GTK_STATUSBAR (gdict->statusbar), 0, _("No matches found"));
+  word = gdict_defbox_get_word (gdict->defbox);
+  gdict_lookup_spelling (gdict, word, FALSE);
+}
+
+static void
+gdict_defbox_substr_not_found_cb (GtkWidget *widget, gpointer data) 
+{
+  GDictWindow *gdict = data;
+  gtk_statusbar_pop (GTK_STATUSBAR (gdict->statusbar), 0);
+  gtk_statusbar_push (GTK_STATUSBAR (gdict->statusbar), 0, _("String not found"));
+}
+
+static void
+gdict_defbox_follow_if_link (GtkWidget *defbox, GtkTextIter *iter, GDictWindow *gdict, int mousebutton)
+{
+  GSList *tags = NULL, *tagp = NULL;
+  gchar *str;
+
+  tags = gtk_text_iter_get_tags (iter);
+  for (tagp = tags;  tagp != NULL;  tagp = tagp->next) {
+    GtkTextTag *tag = tagp->data;
+    str = g_object_get_data (G_OBJECT (tag), "page");
+
+    if (mousebutton == 2)
+      gdict = GDICT_WINDOW (gdict_window_new());
+    else
+      gdict_defbox_clear (GDICT_DEFBOX (defbox));
+
+    gdict_lookup_definition (gdict, str);
+    gtk_widget_grab_focus (gdict->word_entry);
+    gtk_editable_set_position (GTK_EDITABLE (gdict->word_entry), -1);
+    g_free (str);
+  }
+
+  if (tags) 
+    g_slist_free (tags);
+}
+
+static void
+gdict_defbox_set_cursor_if_appropriate (GDictWindow *gdict, GtkWidget *defbox, gint x, gint y)
+{
+  GSList *tags = NULL, *tagp = NULL;
+  GtkTextBuffer *buffer;
+  GtkTextIter iter;
+  static gboolean hovering_over_link = FALSE;
+  static GdkCursor *hand_cursor = NULL;
+  static GdkCursor *regular_cursor = NULL;
+  gboolean hovering = FALSE;
+
+  if (hand_cursor == NULL)
+    hand_cursor = gdk_cursor_new (GDK_HAND2);
+  if (regular_cursor == NULL)
+    regular_cursor = gdk_cursor_new (GDK_XTERM);
+
+  buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (defbox));
+
+  gtk_text_view_get_iter_at_location (GTK_TEXT_VIEW (defbox), &iter, x, y);
+
+  tags = gtk_text_iter_get_tags (&iter);
+  for (tagp = tags;  tagp != NULL;  tagp = tagp->next) {
+    GtkTextTag *tag = tagp->data;
+    gint page = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (tag), "page"));
+
+    if (page != 0) {
+      hovering = TRUE;
+      break;
+    }
+  }
+
+  if (hovering != hovering_over_link) {
+    hovering_over_link = hovering;
+
+    if (hovering_over_link)
+      gdk_window_set_cursor (gtk_text_view_get_window (GTK_TEXT_VIEW (defbox),
+                 GTK_TEXT_WINDOW_TEXT), hand_cursor);
+    else
+      gdk_window_set_cursor (gtk_text_view_get_window (GTK_TEXT_VIEW (defbox),
+                 GTK_TEXT_WINDOW_TEXT), regular_cursor);
+  }
+
+  if (tags)
+    g_slist_free (tags);
+}
+
+static gboolean
+gdict_defbox_event_after (GtkWidget *defbox, GdkEvent *ev, gpointer data)
+{
+  GDictWindow *gdict = data;
+  GtkTextIter iter;
+  GtkTextBuffer *buffer;
+  GdkEventButton *event;
+  gint x, y;
+
+  if (ev->type != GDK_BUTTON_RELEASE)
+    return FALSE;
+
+  event = (GdkEventButton *)ev;
+  if (event->button != 1 && event->button !=2)
+    return FALSE;
+
+  buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (defbox));
+
+  if (gtk_text_buffer_get_selection_bounds (buffer, NULL, NULL))
+    return FALSE;
+
+  gtk_text_view_window_to_buffer_coords (GTK_TEXT_VIEW (defbox), 
+                 GTK_TEXT_WINDOW_WIDGET,
+                 event->x, event->y, &x, &y);
+
+  gtk_text_view_get_iter_at_location (GTK_TEXT_VIEW (defbox), &iter, x, y);
+
+  gdict_defbox_follow_if_link (defbox, &iter, gdict, event->button);
+
+  return FALSE;
+}
+
+static gboolean
+gdict_defbox_motion_notify_event (GtkWidget *defbox,
+                     GdkEventMotion *event, gpointer data)
+{
+  GDictWindow *gdict = data;
+  gint x, y;
+
+  gtk_text_view_window_to_buffer_coords (GTK_TEXT_VIEW (defbox), 
+                 GTK_TEXT_WINDOW_WIDGET,
+                 event->x, event->y, &x, &y);
+  
+  gdict_defbox_set_cursor_if_appropriate (gdict, defbox, x, y);
+  gdk_window_get_pointer (GTK_WIDGET(defbox)->window, NULL, NULL, NULL);
+
+  return FALSE;
+}
+
+/* update the cursor image if the window becomes visible */
+static gboolean
+gdict_defbox_visibility_notify_event (GtkWidget *defbox,
+                         GdkEventVisibility *event, gpointer data)
+{
+  GDictWindow *gdict = data;
+  gint wx, wy, bx, by;
+
+  gdk_window_get_pointer (GTK_WIDGET(defbox)->window, &wx, &wy, NULL);
+
+  gtk_text_view_window_to_buffer_coords (GTK_TEXT_VIEW (defbox), 
+                 GTK_TEXT_WINDOW_WIDGET,
+                 wx, wy, &bx, &by);
+
+  gdict_defbox_set_cursor_if_appropriate (gdict, defbox, bx, by);
+
+  return FALSE;
+}
 
 GType 
 gdict_defbox_get_type (void)
@@ -134,10 +344,9 @@ gdict_defbox_class_init (GDictDefboxClass *class)
                       G_SIGNAL_RUN_FIRST,
                       G_STRUCT_OFFSET (GDictDefboxClass, socket_error),
                       NULL, NULL,
-                      g_cclosure_marshal_VOID__VOID,
+                      g_cclosure_marshal_VOID__STRING,
                       G_TYPE_NONE,
-                      1,
-                      G_TYPE_STRING);
+                      1, G_TYPE_STRING);
 
     class->word_lookup_start = NULL;
     class->word_lookup_done = NULL;
@@ -151,28 +360,32 @@ gdict_defbox_new (gpointer gdict)
     GDictDefbox *defbox;
 
     defbox = GDICT_DEFBOX (g_object_new (GDICT_TYPE_DEFBOX, NULL));
+    gtk_text_view_set_left_margin (GTK_TEXT_VIEW (defbox), 4);
+    gtk_text_view_set_editable (GTK_TEXT_VIEW (defbox), FALSE);
+
     defbox->gdict = gdict;
+    defbox->context = GDICT_WINDOW(gdict)->context;
+    gdict_defbox_setup_tags (defbox);  
+    g_signal_connect (G_OBJECT(defbox), "event-after", 
+                      G_CALLBACK (gdict_defbox_event_after), gdict);
+    g_signal_connect (G_OBJECT(defbox), "motion-notify-event", 
+                      G_CALLBACK (gdict_defbox_motion_notify_event), gdict);
+    g_signal_connect (G_OBJECT(defbox), "visibility-notify-event", 
+                      G_CALLBACK (gdict_defbox_visibility_notify_event), gdict);
+    g_signal_connect (G_OBJECT (defbox), "word_lookup_start",
+                      G_CALLBACK (gdict_defbox_lookup_start_cb), gdict);
+    g_signal_connect (G_OBJECT (defbox), "word_lookup_done",
+                      G_CALLBACK (gdict_defbox_lookup_done_cb), gdict);
+    g_signal_connect (G_OBJECT (defbox), "word_not_found",
+                      G_CALLBACK (gdict_defbox_not_found_cb), gdict);
+    g_signal_connect (G_OBJECT (defbox), "substr_not_found",
+                      G_CALLBACK (gdict_defbox_substr_not_found_cb), gdict);
+    g_signal_connect (G_OBJECT (defbox), "socket_error",
+                      G_CALLBACK (gdict_socket_error_cb), gdict);
 
     return GTK_WIDGET (defbox);
 }
 
-/* Setup textview tags
- */
-void
-defbox_setup_tags (GDictDefbox *defbox)
-{
-    GtkTextBuffer *buffer;
-	
-    buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (defbox));
-	
-    gtk_text_buffer_create_tag (buffer, "italic",
-			        "style", PANGO_STYLE_ITALIC, NULL);
-    gtk_text_buffer_create_tag (buffer, "bold",
-			        "weight", PANGO_WEIGHT_BOLD, NULL);
-    gtk_text_buffer_create_tag (buffer, "underline",
-			        "underline", PANGO_UNDERLINE_SINGLE, NULL);
-}
-
 /* gdict_defbox_lookup
  *
  * Sends the command to the server to commence looking up the definition
@@ -206,9 +419,9 @@ gdict_defbox_lookup (GDictDefbox *defbox, gchar *text)
     defbox->database = g_strdup (gdict->pref->database);
     
     defbox->def_cmd = dict_define_command_new (defbox->database, text);
-    defbox->def_cmd->error_notify_cb = def_error_cb;
-    defbox->def_cmd->status_notify_cb = def_status_cb;
-    defbox->def_cmd->data_notify_cb = def_data_cb;
+    defbox->def_cmd->error_notify_cb = gdict_defbox_error_cb;
+    defbox->def_cmd->status_notify_cb = gdict_defbox_status_cb;
+    defbox->def_cmd->data_notify_cb = gdict_defbox_data_cb;
     defbox->def_cmd->user_data = defbox;
 
     if (dict_command_invoke (defbox->def_cmd, defbox->context) < 0)
@@ -217,11 +430,6 @@ gdict_defbox_lookup (GDictDefbox *defbox, gchar *text)
     return 0;
 }
 
-/* gdict_defbox_clear
- *
- * Clears the text in a defbox and eliminates the current command structure
- */
-
 void 
 gdict_defbox_clear (GDictDefbox *defbox)
 {
@@ -242,7 +450,6 @@ gdict_defbox_clear (GDictDefbox *defbox)
 }
 
 /* gdict_defbox_find
- *
  * Finds a string of text in the current definition
  */
 
@@ -333,7 +540,7 @@ gdict_defbox_print_preview (GDictDefbox *defbox)
 	GtkWidget *preview;
 	GDictWindow *gdict = defbox->gdict;
 	preview = gnome_print_job_preview_new (gdict->print_job, _("Definition preview"));
-	gtk_widget_show (preview);
+	gtk_window_present (GTK_WINDOW(preview));
 }
 
 /* gdict_defbox_print
@@ -348,10 +555,11 @@ gdict_defbox_prepare_print (GDictDefbox *defbox)
     GnomePrintContext *pc;
     GList *node;
     dict_res_t *res;
-    int nb_node = 0, i=0, width, height;
+    int nb_node = 0, i=0, width, height, row_size;
+    int nb_lines, nb_pages, nb_lines_per_page, line_to_print, x, y;
     double page_width, page_height;
     gchar **node_definition;
-    gchar *text;
+    gchar *text, **textlines;
     PangoLayout *layout;
     PangoFontDescription *desc;
 
@@ -361,37 +569,57 @@ gdict_defbox_prepare_print (GDictDefbox *defbox)
     if (defbox->def_cmd == NULL)
 	    return;
 
-    pc = gnome_print_job_get_context (gdict->print_job);
-    g_return_if_fail(pc != NULL);
-    gnome_print_beginpage (pc, "1");
-    gnome_print_job_get_page_size (gdict->print_job, &page_width, &page_height);
-    gnome_print_moveto (pc, 100, page_height - TOP_MARGIN);
-
-    layout = gnome_print_pango_create_layout (pc);
-    pango_layout_set_width (layout, (page_width - LEFT_MARGIN - RIGHT_MARGIN) * PANGO_SCALE);
-    desc = pango_font_description_from_string ("Times 10");
-    pango_layout_set_font_description (layout, desc);
-    pango_font_description_free (desc);
-
-    pango_layout_get_size (layout, &width, &height);
-
     nb_node = g_list_length (defbox->def_cmd->res_list);
-
     node_definition = (gchar **) g_new0 (gpointer, nb_node+1);
     for (node = defbox->def_cmd->res_list; node; node = g_list_next (node)) {
 	    res = (dict_res_t *) node->data;
 	    node_definition[i++] = res->desc;
     }
     node_definition[i] = NULL;
-
     text = g_strjoinv ("\n", node_definition);
+    textlines = g_strsplit (text, "\n", -1);    
+    for (i=0; textlines[i] != NULL; i++);
+    nb_lines = i;
+    
+    row_size = 10;
 
-    pango_layout_set_text (layout, text, -1);
-    gnome_print_pango_layout (pc, layout);
-	    
-    gnome_print_showpage (pc);
+    pc = gnome_print_job_get_context (gdict->print_job);
+    g_return_if_fail(pc != NULL);
+    layout = gnome_print_pango_create_layout (pc);
+    desc = pango_font_description_from_string ("Times 10");
+    pango_layout_set_font_description (layout, desc);
+    pango_font_description_free (desc);
+    pango_layout_set_width (layout, (page_width - LEFT_MARGIN - RIGHT_MARGIN) * PANGO_SCALE);
 
+    gnome_print_job_get_page_size (gdict->print_job, &page_width, &page_height);
+    nb_lines_per_page = ((page_height - TOP_MARGIN*2.)/row_size);
+    nb_pages = ceil ((float) nb_lines / (float) nb_lines_per_page);
+
+    line_to_print = 0;
+    x = 100;
+    for (i=0; i<nb_pages; i++) {
+      int j;
+      gchar *page_name;
+
+      page_name = g_strdup_printf("%d", i+1);
+      gnome_print_beginpage (pc, page_name);
+      g_free (page_name);
+      
+      y = page_height - TOP_MARGIN;
+      for (j=0; j < nb_lines_per_page; j++) {
+        if (line_to_print >= nb_lines)
+          break;
+        gnome_print_moveto (pc, x, y);      
+        pango_layout_set_text (layout, textlines[line_to_print++], -1);
+        gnome_print_pango_layout (pc, layout);
+        y -= row_size;
+      }
+      gnome_print_showpage (pc);
+      if (line_to_print >= nb_lines)
+        break;
+    }
     g_object_unref (G_OBJECT (pc));
+    gnome_print_job_close (gdict->print_job);
 }
 
 /* gdict_defbox_get_word
@@ -508,7 +736,7 @@ insert_link (GtkTextBuffer *buffer, GtkTextIter *iter, gchar *text, gint len)
  */
 
 static void 
-defbox_add (GDictDefbox *defbox, gchar *def)
+gdict_defbox_add (GDictDefbox *defbox, gchar *def)
 {
     GtkTextBuffer *buffer;
     GtkTextIter iter;
@@ -651,30 +879,30 @@ defbox_add (GDictDefbox *defbox, gchar *def)
    
 }
 
-/* def_error_cb
+/* gdict_defbox_error_cb
  *
  * Callback invoked when there was an error in the last query
  */
 
 static void
-def_error_cb (dict_command_t *command, DictStatusCode code,
+gdict_defbox_error_cb (dict_command_t *command, DictStatusCode code,
               gchar *message, gpointer data)
 {
     GObject *defbox;
     dict_command_t *sec_cmd;
-
+    
     defbox = G_OBJECT (data);
-
+    
     if (code != DICT_SOCKET_ERROR) {
         GtkWidget *dialog;
         
         dialog = gtk_message_dialog_new_with_markup (NULL, GTK_DIALOG_DESTROY_WITH_PARENT,
-						     GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
-						     "<span weight=\"bold\" size=\"larger\">%s</span>\n\n%s",
-						     _("Error invoking query"), message);
-	gtk_dialog_run (GTK_DIALOG (dialog));
-	gtk_widget_destroy (dialog);
-	
+                                                     GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
+                                                     "<span weight=\"bold\" size=\"larger\">%s</span>\n\n%s",
+                                                     _("Error invoking query"), message);
+        gtk_dialog_run (GTK_DIALOG (dialog));
+        gtk_widget_destroy (dialog);
+        
         sec_cmd = dict_disconnect_command_new ();
         dict_command_invoke (sec_cmd, command->context);
         
@@ -685,27 +913,27 @@ def_error_cb (dict_command_t *command, DictStatusCode code,
     }
 }
 
-/* def_data_cb
+/* gdict_defbox_data_cb
  *
  * Callback used when a new definition has arrived over the link
  */
 
 static void 
-def_data_cb (dict_command_t *command, dict_res_t *res, gpointer data)
+gdict_defbox_data_cb (dict_command_t *command, dict_res_t *res, gpointer data)
 {
     GDictDefbox *defbox;
     
     defbox = GDICT_DEFBOX (data);
-    defbox_add (defbox, res->desc);
+    gdict_defbox_add (defbox, res->desc);
 }
 
-/* def_status_cb
+/* gdict_defbox_status_cb
  *
  * Callback used when a status code has arrived over the link
  */
 
 static void 
-def_status_cb (dict_command_t *command, DictStatusCode code, 
+gdict_defbox_status_cb (dict_command_t *command, DictStatusCode code, 
                int num_found, gpointer data)
 {
     GObject *defbox;
